/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of ADORB.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#import <Foundation/Foundation.h>
#import "IDLDefinition_Private.h"
#import "IDLEnum.h"
#import "IDLEncapsulation.h"
#import <objc/runtime.h>

@implementation IDLEnum(CORBACoding)

NSString* CORBAInvalidArgumentException = @"CORBAInvalidArgumentException";

- (void) encodeObject:obj withCoder:(id <CORBACoder>)inv
{
    [inv encodeULong:[obj unsignedLongValue]];
}

- (void) decodeObject:(id*)obj withCoder:(id <CORBADecoder>)inv
{
    unsigned long l;
    [inv decodeULong:&l];
    *obj = [NSNumber numberWithUnsignedLong:l];
}

- (const char*) objCType
{
    return @encode(long);
}
- (unsigned) argLength
{
    return sizeof(long);
}
- (unsigned) returnValueLength
{
    return sizeof(long);
}

- argumentObjectFromInvocation:(NSInvocation*)inv atIndex:(int)n
{
    long value = 0;
    [inv getArgument:&value atIndex:n];
    return [NSNumber numberWithInt:value];
}

- argumentObjectOfType:(const char*)argType fromPointer:(void*)ptr
{
    if(argType[0] == _C_PTR){
        switch(argType[1]){
            case _C_LNG:
            case _C_ULNG:{
                long* value = ptr;
                return [NSNumber numberWithLong:*value];
            }
            break;
            case _C_ID:{
                id * value = ptr;
                if(!*value || [*value isKindOfClass:[NSNumber class]])
                    return *value;
            }
        }
    }
    [NSException raise:CORBAInvalidArgumentException
        format:@"invalid argument type '%s', must be 'long*' or compatible", argType];
    return nil;
}

- (void) setValue:value forInvocation:(NSInvocation*)inv atIndex:(int)n
{
    const char* argType = [[inv methodSignature]getArgumentTypeAtIndex:n];
    switch(argType[0]){
        case _C_LNG:
        case _C_ULNG:{
            long _val = [value intValue];
            [inv setArgument:&_val atIndex:n];
        }
        break;
        case _C_ID:{
            id _val = value;
            [inv setArgument:&_val atIndex:n];
        }
        break;
        default:
            [NSException raise:CORBAInvalidArgumentException
                format:@"invalid argument type '%s', must be 'long' or compatible", argType];
    }
}

- (void) setValue:value ofType:(const char*)argType forPointer:(void*)ptr
{
    switch(argType[1]){
        case _C_LNG:
        case _C_ULNG:{
            long _val = [value intValue];
            *((long*)ptr) = _val;
        }
        break;
        case _C_ID:
            *((id*)ptr) = value;
            break;
        default:
            [NSException raise:CORBAInvalidArgumentException 
                format:@"invalid argument type '%s', must be 'long*' or compatible", argType];
    }
}
- (void) setReturnValue:value forInvocation:(NSInvocation*)inv
{
    long _val = [value intValue];
    [inv setReturnValue:&_val];
}

- returnValueFromInvocation:(NSInvocation*)inv
{
	// FixMe: does not check if the value is valid
	const char * retType = [[inv methodSignature]methodReturnType];
	switch(retType[0]){
		case _C_LNG:
		case _C_ULNG: {
			long _val = 0;
			[inv getReturnValue:&_val];
			return [NSNumber numberWithLong:_val];
		}
		case _C_ID: {
			id _val = nil;
			[inv getReturnValue:&_val];
			return _val;
		}
		default:
            [NSException raise:CORBAInvalidArgumentException 
                format:@"invalid return type '%s', must be 'long' or compatible", retType];
	}
	return nil;
}

- (void) _encodeWithCORBACoder:(id <CORBACoder>)coder
{
    int count = [_enumNames count];
    int j;
    NSEnumerator* en = [_enumNames keyEnumerator];
    NSString* key;
    NSString* names[count];
    [coder encodeString:[self typeId]];
    [coder encodeString:[self name]];
    [coder encodeLong:count];
    for(j=0;j<count;j++)
        names[j] = @"undefined";
    while(key = [en nextObject]){
        int n = [[_enumNames objectForKey:key]intValue];
        if(n < count)
            names[n] = key;
    }
    for(j=0;j<count;j++)
        [coder encodeString:names[j]];
}

- (void) encodeWithCORBACoder:(id <CORBACoder>)coder
{
    long location = [coder locationForDefinition:self];
    if(location != 0 && [IDLDefinition allowsIndirection]) // already encoded -> encode indirection
        [coder encodeIndirection:location];
    else{
        IDLEncapsulation* stream = [IDLEncapsulation encapsulation];
        [coder encodeLong:[self kind]];
        [stream setParentStream:coder location:[coder location]];
        [stream registerLocation:-2*sizeof(long) forDefinition:self];
        [self _encodeWithCORBACoder:stream];
        [coder encodeData:[stream data]];
    }
}

+ _definitionFromCoder:(IDLEncapsulation*)coder
{
    NSMutableDictionary* dict = [NSMutableDictionary dictionary];
    NSString *str;
    NSString* typeId;
    IDLEnum* obj;
    long cnt = 0;
    long n = 0;
    
    [coder decodeString:&typeId];
    [coder decodeString:&str];
    [coder decodeLong:&cnt];
    
    while(cnt-- > 0){
        NSString* member;
        [coder decodeString:&member];
        [dict setObject:[NSNumber numberWithLong:n++] forKey:member];
    }
    
    obj = [self enumWithDictionary:dict name:str];
    [obj setTypeId:typeId];
    [coder registerDefinition:obj forLocation:-2*sizeof(long)];
    return obj;
}

@end
