/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of ADORB.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#import "VXMessageHandler.h"
#import "VXMessage.h"
#import "VXConnection.h"
#import "CORBAObject_Private.h"
#import "IDLDefinitionScope_Private.h"
#import "CORBAException.h"
#import "ORB_Private.h"
#import <IDL.h>
#import "IDLIOPProfile.h"
#import "IDLProfileBody.h"
#import "IDLGIOPVersion.h"
#import "IDLServiceContext.h"

@interface IDLServiceContext(Handler)
+ (void) setCurrentServiceContext:(NSDictionary*)ctxt;
@end

@implementation VXMessageHandler

static BOOL _trace = NO;
static BOOL _debug = NO;

+ (void) initialize
{
    if([VXMessageHandler class] == self){
        NSUserDefaults* defaults = [NSUserDefaults standardUserDefaults];
        NSString* str = [defaults stringForKey:@"debug"];
        _trace = [defaults boolForKey:@"trace"];
        if(str && [str rangeOfString:@"m"].length > 0)
            _debug = YES;
    }
}
- (void) invalidateConnection:(NSNotification*)notification
{
    [_connection release];
    _connection = nil;
}

- initWithMessage:(VXMessage*)msg connection:(VXConnection*)conn
{
	self = [super init];
    _connection = [conn retain];
    [[NSNotificationCenter defaultCenter]addObserver:self
                selector:@selector(invalidateConnection:)
                name:VXInvalidateConnection
                object:_connection];
	_requestContext = [[NSMutableDictionary alloc]init];
	_responseContext = [[NSMutableDictionary alloc]init];
	_flowStack = [[NSMutableArray alloc]init];
	@try{
		_message = [msg retain];
		_completionStatus = COMPLETED_NO;
		[_message setInvocationObject:self];
		_msgType = [_message messageType];
		[self decodeRequestHeader];
    }@catch(NSException *localException){
		_exception = [localException retain];
		if(_debug)
			NSLog(@"%@: %@\n",[localException name], [localException reason]);
    }
    return self;
}

- (void) dealloc
{
    [[NSNotificationCenter defaultCenter]removeObserver:self];
    [_connection release];
    [_message release];
    [_requestHeader release];
    [_exception release];
    [_target release];
    [_targetObject release];
	[_requestContext release];
	[_responseContext release];
	[_flowStack release];
    [super dealloc];
}

+ handlerWithMessage:(VXMessage*)msg connection:(VXConnection*)conn
{
	return [[[self alloc]initWithMessage:msg connection:conn]autorelease];
}

- connection
{
	return _connection;
}

- (void) _addServiceContexts:(NSArray*)array
{
	NSEnumerator* en = [array objectEnumerator];
	IDLServiceContext* ctxt;
	while(ctxt = [en nextObject]){
		[_requestContext setObject:ctxt forKey:[NSNumber numberWithInt:[ctxt contextId]]];
		switch([ctxt contextId]){
			case ORB_CodeSets:
				[_connection setStringEncodingsFromCodeSets:ctxt];
				[[_message stream]setStringEncoding:[_connection stringEncoding]];
				[[_message stream]setWStringEncoding:[_connection wstringEncoding]];
				break;
			
			case ORB_BI_DIR_IIOP: // path = @"context.listen_points"
				[_connection addListenPoints:[ctxt valueForKeyPath:@"context.listen_points"]];
				break;
			default:
				break;
		}
	}
}

- (void) decodeRequestHeader
{
    unsigned major = [_message majorVersion];
    unsigned minor = [_message minorVersion];
    NSString* path = nil;
    IDLDefinition* def;
    NSData* objectKey;
    
    switch(_msgType){
        case GIOP_RequestType:
            path = [NSString stringWithFormat:@"GIOP/RequestHeader_%d_%d", major, minor];
            break;
        case GIOP_LocateRequestType:
            path = [NSString stringWithFormat:@"GIOP/LocateRequestHeader_%d_%d", major, minor];
            break;
        case GIOP_CancelRequestType:
            path = @"GIOP/CancelRequestHeader";
            break;
    }
       
	def = (IDLDefinition*)[[IDLDefinitionScope rootScope]definitionForNamePath:path];
    if(def){
        [def decodeObject:&_requestHeader withCoder:[_message stream]];
        [_requestHeader retain];
		[self _addServiceContexts:[_requestHeader valueForKey:@"service_context"]];
    }else
        [CORBAException raise:@"INITIALIZE" minorCode:GIOPDefinitionError completed:CORBA_CompletedNo];
        //[NSException raise:CORBADefinitionException format:@"No definition for %@",path];

    switch(minor){
        case 0:
        case 1:
            if(objectKey = [_requestHeader valueForKey:@"object_key"])
                _targetObject = [[self locateObjectWithObjectID:objectKey]retain];
            _target = [[_targetObject implementation]retain];
            _targetInterface = [_targetObject interface];
            break;
        case 2:{
            NSDictionary* targetDict = [_requestHeader valueForKey:@"target"];
            switch([[_target valueForKey:IDLUnionDescriptorKey]shortValue]){
                case 0:
                    if(objectKey = [targetDict valueForKey:IDLUnionDataKey])
                        _targetObject = [[self locateObjectWithObjectID:objectKey]retain];
                    _target = [[_targetObject implementation]retain];
                    _targetInterface = [_targetObject interface];
                    break;
                case 1:{
                    IDLIOPProfile* profile = [targetDict valueForKey:IDLUnionDataKey]; //TagedProfile
                    unsigned long tag = [profile tag];
                    if(tag == TAG_INTERNET_IOP){
                        IDLProfileBody* body = [profile profileBody];
                        // Don't use components yet.
                        //NSArray* components = [body components];
                        //NSString* host = [body hostName];
                        //unsigned port = [body portNumber];
                        objectKey = [body objectID];
#warning FIXME: Does not check host & port in RequestHeader target (case 1)
                        if(objectKey)
                            _targetObject = [[self locateObjectWithObjectID:objectKey]retain];
                        _target = [[_targetObject implementation]retain];
                        _targetInterface = [_targetObject interface];
                        
                    }else{
                        NSLog(@"GIOP %d.%d: RequestHeader/ProfileAddr - ProfileId '%d' is not supported",
                            major,minor, tag);
                        [CORBAException raise:@"NO_IMPLEMENT" 
                            minorCode:ProfileIdNotSupported completed:CORBA_CompletedNo];
                        //[NSException raise:@"CORBAImplementationException"
                        //    format:@"GIOP %d.%d: RequestHeader/ProfileAddr - ProfileId '%d' is not supported",
                        //        major,minor, tag];
                    }
                }
                break;
                
                case 2:{
#warning FIXME: GIOP 1.2, RequestHeader target (case 2) not fully implemented
                    // value is a dictionary with 'lon' and 'CORBAObject'
                    // The CORBAObject has different profile and does not match local object!
                    NSDictionary* addrInfo = [targetDict valueForKey:IDLUnionDataKey]; 
                    //unsigned long i = [[addrInfo valueForKey:@"selected_profile_index"]unsignedLongValue];
                    _targetObject = [[addrInfo valueForKey:@"ior"]retain]; // CORBAObject
                    if([_targetObject isKindOfClass:[CORBAObject class]]){
                        if([_targetObject isLocal]){
                            _target = [[_targetObject implementation]retain];
                            _targetInterface = [_targetObject interface];
                        }else{
                            // Either iorObj has multiple TAG_INTERNET_IOP profiles or it is not ours
                            //NSArray* profiles = [_targetObject profiles];
                            NSLog(@"CORBAInvalidObjectReference: GIOP %d.%d: RequestHeader/ReferenceAddr - cannot locate object."
								@"Either the object is not here or 'ReferenceAddr' has multiple TAG_INTERNET_IOP profiles, which is not supported", 
                                    major,minor);
                        
                            [CORBAException raise:@"INV_OBJREF"
                                minorCode:ProfileError completed:NO];
                        }
                    }else{ // servant
                        _target = _targetObject;
                        _targetObject = [[CORBAObject lookupImplementation:_target]retain];
                        _targetInterface = [_targetObject interface];
                    }
                }
                break; 
            }
        }
    }
}

- (void) _appendReplyHeader:(VXMessage*)message
{
    NSString* path = [NSString stringWithFormat:@"GIOP/ReplyHeader_%d_%d",
            [message majorVersion], [message minorVersion]];
    IDLDefinition* def = (IDLDefinition*)[[IDLDefinitionScope rootScope]definitionForNamePath:path];
    if(def){
        [[message stream] encodeStruct:self usingObjectCoder:def];
    }else
        [CORBAException raise:@"INITIALIZE" 
            minorCode:GIOPDefinitionError completed:CORBA_CompletedYes];
        //[NSException raise:CORBADefinitionException format:@"cannot find definition for %@",path];
}

- locateObjectWithObjectID:(NSData*)objectKey
{
    CORBAObject* obj = [_connection objectForObjectID:objectKey];
    if(obj)
        return obj;
    [CORBAException raise:@"INV_OBJREF" minorCode:ObjectNotFound completed:CORBA_CompletedNo];
    //[NSException raise:@"CORBAInvalidObjectReference"
    //           format:@"unable to locate object with objectKey '%@'", objectKey];
    return nil;
}

- locateObject
{
    NSData* objectKey = [_requestHeader valueForKey:@"object_key"];
    return [self locateObjectWithObjectID:objectKey];
}

- (void) _sendException:(NSException*)exception
{
    NSDictionary* info = [CORBAException standardExceptionInfoForName:[exception name]];
    VXMessage* reply = [[[VXMessage allocWithZone:[self zone]]
            initWithMessageType:GIOP_ReplyType
            iiopVersion:[_message iiopVersion]]autorelease];
	[reply setInvocationObject:self];

    
	IDLDefinition* def = nil;
    if(info){
        def = (IDLDefinition*)[[IDLDefinitionScope rootScope]definitionForNamePath:
                @"GIOP/SystemExceptionReplyBody"];
		_replyStatus = GIOP_SystemException;
    }else{
        IDLDefinitionScope* scope = _targetInterface?[_targetInterface rootScope] : [IDLDefinitionScope rootScope];
        info = [exception userInfo];
		NSNumber* excType = [info valueForKey:IDLExceptionTypeKey];
		_replyStatus = excType? [excType unsignedIntValue] : GIOP_UserException;
		def = [info valueForKey:IDLExceptionKey];
		if(!def){
			NSString *typeId = [info valueForKey:IDLExceptionIDKey];
			if(typeId)
				def = (IDLDefinition*)[scope definitionForTypeId:typeId];
			else
				def = (IDLDefinition*)[_targetInterface definitionNamed:[exception name]];
			if(!def){// Unrecognized exception, send system 'UNKNOWN'
				def = (IDLDefinition*)[[IDLDefinitionScope rootScope]definitionForNamePath:
							@"GIOP/SystemExceptionReplyBody"];
				info = [CORBAException standardExceptionInfoForName:@"CORBAUnknownException"];
				_replyStatus = GIOP_SystemException;
			}
		}
    }

    [self _appendReplyHeader:reply];
    [reply alignStreamBody];
	[def encodeObject:info withCoder:[reply stream]];
    [_connection sendMessage:reply];
}

- (void) _handleException:(NSException*)exception
{
	while([_flowStack count] > 0){
		NSDictionary* replyInfo = [NSDictionary dictionaryWithObjectsAndKeys:
			_targetObject, ORBObjectKey,
			[_function name], ORBOperationNameKey,
			_responseContext, ORBServiceContextKey,
			exception, ORBExceptionKey,
			[_requestHeader valueForKey:@"request_id"], ORBRequestIdKey,
			[NSNumber numberWithBool:[_function isOneWay]], ORBOnewayKey,
			nil];
		if([CORBAException isForwardRequestException:exception])
			exception = [[ORB notificationCenter]postWillSendOther:self userInfo:replyInfo flowStack:_flowStack];
		else
			exception = [[ORB notificationCenter]postWillSendException:self userInfo:replyInfo flowStack:_flowStack];
	}
	if([self responseExpected] && _connection)
		[self _sendException:exception];
}

- (void) _handleRequest
{
	id effectiveTarget = nil;
	NSException* exception = nil;
    NSString* opName = [_requestHeader valueForKey:@"operation"];
    if(_trace)
        NSLog(@"PROCESSING REQUEST <%x> '%@' from '%@' (IIOP%d.%d)\n",
                [[self request_id]unsignedLongValue], opName,
                [(VXConnection*)_connection address],
                [_message majorVersion], [_message minorVersion]);

	// Call interceptor's point for request service context
	NSDictionary* info = [NSDictionary dictionaryWithObjectsAndKeys:
			_targetObject, ORBObjectKey,
			[_function name], ORBOperationNameKey,
			_requestContext, ORBServiceContextKey,
			[_requestHeader valueForKey:@"request_id"], ORBRequestIdKey,
			nil];
	exception = [[ORB notificationCenter]postDidReceiveRequestContext:self userInfo:info flowStack:_flowStack];
    if(exception){
		[self _handleException:exception];
		return;
	}

    @try{
		[[_targetInterface rootScope]set]; // set current scope
		if(_debug)
			NSLog(@"target: %@ - %@ interface: %@", 
				[_targetObject description], [_target description], [_targetInterface typeId]);
		
		_function = [_targetInterface functionNamed:opName];
		// post notification allowing to check service context
		[IDLServiceContext setCurrentServiceContext:[NSDictionary dictionaryWithDictionary:_requestContext]];
		
		[_message alignStreamBody];
		if(_function){
			// Function may provide two different selectors, old style and new style
			// Try both selectors
			SEL sel = NSSelectorFromString([_function selectorName]);
			NSInvocation* inv = nil;
			[_function decodeRequest:&_args withCoder:[_message stream]];
			if([_target respondsToSelector:sel]){ // Implementation can override Object methods
				NSMethodSignature* ms = [_target methodSignatureForSelector:sel];
				inv = [_function invocationWithArguments:_args methodSignature:ms];
				[inv setSelector:_selector = sel];
				effectiveTarget = _target;
			}else
			if([_targetObject respondsToSelector:sel]){ // try CORBAObject itself
				NSMethodSignature* ms = [_targetObject methodSignatureForSelector:sel];
				inv = [_function invocationWithArguments:_args methodSignature:ms];
				[inv setSelector:_selector = sel];
				effectiveTarget = _targetObject;
			}else{
				[CORBAException raise:@"NO_IMPLEMENT" minorCode:TargetDoesNotRespond
					completed:CORBA_CompletedNo];
			}

			// Call interceptor's point for request
			NSDictionary* info = [NSDictionary dictionaryWithObjectsAndKeys:
					_targetObject, ORBObjectKey,
					[_function name], ORBOperationNameKey,
					_requestContext, ORBServiceContextKey,
					[_requestHeader valueForKey:@"request_id"], ORBRequestIdKey,
					[NSNumber numberWithBool:[_function isOneWay]], ORBOnewayKey,
					nil];
			[_flowStack removeAllObjects];
			NSException* exception = [[ORB notificationCenter]postDidReceiveRequest:self userInfo:info flowStack:_flowStack];
			if(exception){
				[self _handleException:exception];
				return;
			}
			[inv invokeWithTarget:effectiveTarget];
			_args = [_function takeValues:_args fromInvocation:inv];
		}else{
			[self _handleException:[CORBAException systemException:@"BAD_OPERATION" minorCode:OperationNotFound completed:CORBA_CompletedNo]];
			return;
		}
    }@catch(NSException *localException){
		exception = [localException retain];
		if(_debug)
			NSLog(@"%@: %@\n",[localException name], [localException reason]);
    }

	if(exception){
		[self _handleException:exception];
		return;
	}

	NSDictionary* replyInfo = [NSDictionary dictionaryWithObjectsAndKeys:
		_targetObject, ORBObjectKey,
		[_function name], ORBOperationNameKey,
		_responseContext, ORBServiceContextKey,
		[_requestHeader valueForKey:@"request_id"],ORBRequestIdKey,
		[NSNumber numberWithBool:[_function isOneWay]], ORBOnewayKey,
		nil];
	_replyStatus = GIOP_NoException;
	// Call interceptor's send reply
	exception = [[ORB notificationCenter] postWillSendResponse:self userInfo:replyInfo flowStack:_flowStack];
	if(exception){
		[self _handleException:exception];
		return;
	}

    if([self responseExpected] && _connection){

		@try{
			VXMessage* reply = [[[VXMessage allocWithZone:[self zone]]
					initWithMessageType:GIOP_ReplyType
					iiopVersion:[_message iiopVersion]]autorelease];
			[reply setInvocationObject:self];
			[self _appendReplyHeader:reply];
			[reply alignStreamBody];
			[_function encodeReply:_args withCoder:[reply stream]];

			[_connection sendMessage:reply];
		}@catch(NSException *localException){
			exception = [[CORBAException systemException:@"MARSHAL" 
					minorCode:MarshalException completed:CORBA_CompletedYes]retain];
			if(_debug)
				NSLog(@"%@: %@\n",[localException name], [localException reason]);
		}

		if(exception)
			[self _handleException:exception];
	}
    if(_trace)
        NSLog(@"FINISHED REQUEST <%x>\n", [[self request_id]unsignedLongValue]);
}

- (void) _handleLocateRequest
{
    NSString* path;
    IDLDefinition* def;
    id locateReplyHeader = [NSMutableDictionary dictionary];
    VXMessage* outMessage;
    long status = (_target != nil)? 1 : 0; //OBJECT_HERE(1) or UNKNOWN_OBJECT(0)
    unsigned minor = [_message minorVersion];
    unsigned major = [_message majorVersion];
    
    @try{
		[locateReplyHeader setValue:[_requestHeader valueForKey:@"request_id"]forKey:@"request_id"];
		[locateReplyHeader setValue:[NSNumber numberWithLong:status]forKey:@"locate_status"];
		outMessage = [[[VXMessage allocWithZone:[self zone]]
					initWithMessageType:GIOP_LocateReplyType 
					iiopVersion:[_message iiopVersion]]autorelease];
		[outMessage setInvocationObject:self];
		path = [NSString stringWithFormat:@"GIOP/LocateReplyHeader_%d_%d", major, minor];
		if(def = (IDLDefinition*)[[IDLDefinitionScope rootScope]definitionForNamePath:path])
			[def encodeObject:locateReplyHeader withCoder:[outMessage stream]];
		else
			[CORBAException raise:@"INITIALIZE" minorCode:GIOPDefinitionError 
					completed:CORBA_CompletedNo];
			//[NSException raise:CORBADefinitionException format:@"No definition for %@",path];
		
		// No reply body for status OBJECT_HERE or UNKNOWN_OBJECT
		[_connection sendMessage:outMessage];
    }@catch(NSException *localException){
		if(_debug)
			NSLog(@"Exception '%@', %@\n",[localException name],[localException reason]);
		[localException raise];
    }
}

- (void) handleMessage
{
    if(!_exception){
        @try{
			switch(_msgType){
				case GIOP_RequestType:
					[self _handleRequest];
					break;
				case GIOP_CancelRequestType:
					// Cancels Request or LocateRequest. Client->Server
					// This is for asynch requests only
					[CORBAException raise:@"IMPL_LIMIT"
						minorCode:AsynchCancelRequest completed:CORBA_CompletedNo];
					//[NSException raise:CORBACommunicationException
					//        format:@"asynchronous CancelRequest is not implemented"];
				case GIOP_LocateRequestType:
					[self _handleLocateRequest];
					break;
				case GIOP_CloseConnectionType:
					[_connection invalidateWithError:[NSError errorWithDomain:CORBAErrorDomain code:kECloseConnection userInfo:nil]];
					break;
				case GIOP_MessageErrorType:
					[CORBAException raise:@"IMPL_LIMIT"
						minorCode:MessageError completed:CORBA_CompletedNo];
				   // [NSException raise:CORBACommunicationException 
				   //     format:@"received 'MessageError' from %@",
				   //         [_connection remoteHostName]];

				default:
					[CORBAException raise:@"IMPL_LIMIT"
						minorCode:UnknownMsgType completed:CORBA_CompletedNo];
				   // [NSException raise:CORBACommunicationException
				   //     format:@"invaid or unsupported message header type %d",
				   //     _msgType];
			}
        }@catch(NSException *localException){
			_exception = [localException retain];
        }
    }
    if(_exception)
        [self _handleException:_exception];
}

- (long unsigned) messageType
{
    return _msgType;
}

- (BOOL) responseExpected
{
    switch([_message minorVersion]){
        case 0:
        case 1:{
			BOOL b;
			[IDLPrimitive setValue:&b fromObject:[_requestHeader valueForKey:@"response_expected"] ofKind:tk_boolean];
			return b;
		}
        //    return [[_requestHeader valueForKey:@"response_expected"]intValue];
        case 2:{
			unsigned char b;
			[IDLPrimitive setValue:&b fromObject:[_requestHeader valueForKey:@"response_flags"] ofKind:tk_octet];
			return b & 0x01;
//            const unsigned char* b = [[_requestHeader valueForKey:@"response_flags"]bytes];
//            return b[0]&0x1;
        }
        default:
            return NO;
    }
}

- (SEL) selector
{
	return _selector;
}

- (NSDictionary*) requestContext
{
	return [NSDictionary dictionaryWithDictionary:_requestContext];
}

- (IDLServiceContext*) requestContextWithContextId:(unsigned)contextId
{
	return [_requestContext objectForKey:[NSNumber numberWithUnsignedInt:contextId]];
}

- (void) addRequestContext:(IDLServiceContext*)context
{
	[_requestContext setObject:context forKey:[NSNumber numberWithUnsignedInt:[context contextId]]];
}

- (void) removeRequestContextWithContextId:(unsigned)contextId
{
	[_requestContext removeObjectForKey:[NSNumber numberWithUnsignedInt:contextId]];
}

- (NSDictionary*) responseContext
{
	return [NSDictionary dictionaryWithDictionary:_responseContext];
}

- (IDLServiceContext*) responseContextWithContextId:(unsigned)contextId
{
	return [_responseContext objectForKey:[NSNumber numberWithUnsignedInt:contextId]];
}

- (void) addResponseContext:(IDLServiceContext*)context
{
	[_responseContext setObject:context forKey:[NSNumber numberWithUnsignedInt:[context contextId]]];
}

- (void) removeResponseContextWithContextId:(unsigned)contextId
{
	[_responseContext removeObjectForKey:[NSNumber numberWithUnsignedInt:contextId]];
}

- (unsigned) completionStatus
{
	return _completionStatus;
}

- (NSDictionary*) arguments
{
	return _args? [NSDictionary dictionaryWithDictionary:_args] : nil;
}

- (IDLFunction*) operationObject
{
	return _function;
}

- (CORBAObject*) object
{
	return _targetObject;
}

#pragma mark ---- KeyValueCoding ----

- (NSArray*) service_context
{
	return [_responseContext allValues];
}

- (NSNumber*) request_id
{
    return [_requestHeader valueForKey:@"request_id"];
}

- (long) requestId
{
	return [[self request_id]longValue];
}

- (unsigned) reply_status
{
	return _replyStatus;
//    return [NSNumber numberWithInt:[self exceptionType]];
}
- (unsigned) replyStatus
{
	return _replyStatus;
//	return [self exceptionType];
}
- (NSStringEncoding) stringEncoding
{
	return [_connection stringEncoding];
}

- (NSStringEncoding) wstringEncoding
{
	return [_connection wstringEncoding];
}

@end
